/* Print out a list of target resolutions and how best to crop/resize to
 * reach the desired aspect ratio.
 *
 * Author: Eric A. Welsh
 * Email:  ewelsh@ccb.wustl.edu
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

struct xy_res
{
    int x, y;
    int x_new, y_new;
    int scaled_crop_left, scaled_crop_right;
    int scaled_crop_top, scaled_crop_bottom;
    float percent_error;
    float crop_x_error, crop_y_error;
    float score;
    int crop_then_resize_flag;
};

int compare_entries1(const void *p1, const void *p2)
{
    struct xy_res *sp1, *sp2;

    sp1 = *((struct xy_res **) p1);
    sp2 = *((struct xy_res **) p2);

    if (sp1->x_new < sp2->x_new)
    	return -1;
    if (sp1->x_new > sp2->x_new)
    	return 1;

    if (sp1->y_new < sp2->y_new)
    	return -1;
    if (sp1->y_new > sp2->y_new)
    	return 1;

    if (sp1->score < sp2->score)
    	return -1;
    if (sp1->score > sp2->score)
    	return 1;
    
    if (fabs(sp1->percent_error) < fabs(sp2->percent_error))
    	return -1;
    if (fabs(sp1->percent_error) > fabs(sp2->percent_error))
    	return 1;

    return 0;
}

int compare_entries2(const void *p1, const void *p2)
{
    struct xy_res *sp1, *sp2;

    sp1 = *((struct xy_res **) p1);
    sp2 = *((struct xy_res **) p2);

    if (sp1->x_new < sp2->x_new)
    	return 1;
    if (sp1->x_new > sp2->x_new)
    	return -1;

    if (sp1->y_new < sp2->y_new)
    	return 1;
    if (sp1->y_new > sp2->y_new)
    	return -1;

    return 0;
}

int main(int argc, char *argv[])
{
    int x_orig, y_orig;
    double aspect_ratio_raw, aspect_ratio_cropped, aspect_ratio_trial;
    int crop_left, crop_right, crop_top, crop_bottom;
    double scaled_crop_left, scaled_crop_right;
    double scaled_crop_top, scaled_crop_bottom;
    double x_img, y_img;
    int final_mod_x = 16, final_mod_y = 16;
    int align_mod_x = 1, align_mod_y = 1;
    int pre_resize_mod_x = 1, pre_resize_mod_y = 1;
    int x_start, y_start;
    double y_scale;
    double temp;

    double percent_error, crop_x_error, crop_y_error, score;
    int x, y, x_mod, y_mod;
    int x_new, y_new;

    unsigned long i;
    struct xy_res **res_array = NULL, **res_array2 = NULL;
    struct xy_res **res_array_cr = NULL, **res_array_cr2 = NULL;
    struct xy_res **res_array_combined = NULL, **res_array_combined2 = NULL;
    long num_entries = 0, num_entries2 = 0;
    long num_entries_cr = 0, num_entries_cr2 = 0;
    long num_entries_combined = 0, num_entries_combined2 = 0;
    long old_x_new = -1, old_y_new = -1;

    if (argc > 7)
    {
    	x_orig = atoi(argv[1]);
    	y_orig = atoi(argv[2]);
    	aspect_ratio_raw = atof(argv[3]);
    	crop_left = atoi(argv[4]);
   	crop_right = atoi(argv[5]);
    	crop_top = atoi(argv[6]);
    	crop_bottom = atoi(argv[7]);
    }
    else
    {
	printf("Program: aspect v1.1\n");
	printf("Author:  Eric A. Welsh <ewelsh@ccb.wustl.edu>\n");
	printf("\n");

	printf("Usage: aspect.exe x_orig y_orig ratio cl cr ct cb [mx my] [amx amy] [prmx prmy]\n");
	printf("Examples: aspect.exe 720 480 1.8228 2 2 59 63 16 16         <RGB space>\n");
	printf("          aspect.exe 720 480 1.8228 2 2 59 63 16 16 2 2     <YV12, progressive>\n");
	printf("          aspect.exe 720 480 1.8228 2 2 59 63 16 16 1 1 1 2 <old VDub Lanczos>\n");
	printf("\n");

#ifdef OLD_ASPECT_RATIOS
	printf("Aspect Ratio to use for each DVD/VCD input resolution:\n");
	printf("     720x480 720x576\n");
	printf("     704x480 704x576 352x480 352x576\n");
	printf("     -------------------------------\n");
	
	/* pixel aspect ratios for the input media */
/*
	 4:3  0.909   1.091   1.818   2.182
	16:9  1.212   1.455   2.424   2.909
*/

	/* display aspect ratios */
	/* pixel aspect ratio * 3/2 for NTSC, 5/4 for PAL */
	printf(" 4:3  1.3635 1.36375  2.7270 2.72750\n");
	printf("16:9  1.8180 1.81875  3.6360 3.63625\n");
#else
	printf("Aspect Ratio to use for each DVD/SVCD/VCD input resolution:\n");
	printf("     720x480 720x576\n");
        printf("     704x480 704x576 480x480 480x576 352x480 352x576\n");
	printf("     -----------------------------------------------\n");

	/* pixel aspect ratios for the input media */
        /* http://www.uwasa.fi/~f76998/video/conversion/ */
	/* multiply 4:3 aspect ratios by 4/3 to get 16:9 anamorphic */
/*
	 4:3  72/79   128/117 108/79  128/78  144/79   256/117
	16:9  288/237 512/351 144/79  256/117 576/237 1024/351

	 4:3  0.9114  1.0940  1.3671  1.6410  1.8228  2.1880
	16:9  1.2152  1.4587  1.8228  2.1880  2.4304  2.9174
*/

	/* display aspect ratios */
	/* pixel aspect ratio * 3/2 for NTSC, 5/4 for PAL */
/*
	 4:3  108/79  160/117 162/79   80/39  216/79   320/117
        16:9  144/79  640/351 216/79  320/117 288/79  1280/351
*/
	printf(" 4:3  1.3671  1.3675  2.0506  2.0513  2.7342  2.7350\n");
        printf("16:9  1.8228  1.8234  2.7342  2.7350  3.6456  3.6467\n");
#endif

	printf("\n");
	printf("x_orig/y_orig = original input resolution\n");
	printf("ratio         = target aspect ratio\n");
	printf("cl/cr/ct/cb   = crop left/right/top/bottom needed to remove black borders\n");
	printf("mx/my         = mods for final X/Y resolutions (default 16/16)\n");
	printf("amx/amy       = alignment mods for X/Y (color space dependant):\n");
	printf("                   RGB:  1/1 (this is the default)\n");
	printf("                   YUY2: 2/1 (progressive), 2/2 (interlaced)\n");
	printf("                   YV12: 2/2 (progressive), 2/4 (interlaced)\n");
	printf("prmx/prmy     = pre-resize mods for X/Y (default 1/1)\n");
	printf("                   ONLY used when an alignment mod == 1\n");
	printf("                   old Lanczos resize VDub plugin requires Y to be mod 2\n");
	exit(1);
    }

    if (argc > 9)
    {
    	final_mod_x = atoi(argv[8]);
    	final_mod_y = atoi(argv[9]);
    }
    else if (argc > 8)
    {
    	printf("Must specify _BOTH_ x and y final resolution mods\n");
    	exit(1);
    }
    if (argc > 11)
    {
    	align_mod_x = atoi(argv[10]);
    	align_mod_y = atoi(argv[11]);
    }
    else if (argc > 10)
    {
    	printf("Must specify _BOTH_ x and y align mods\n");
    	exit(1);
    }
    if (argc > 13)
    {
    	pre_resize_mod_x = atoi(argv[12]);
    	pre_resize_mod_y = atoi(argv[13]);
    }
    else if (argc > 12)
    {
    	printf("Must specify _BOTH_ x and y pre-rescale mods\n");
    	exit(1);
    }
    
    x_new = x_orig - crop_left - crop_right;
    y_new = y_orig - crop_top - crop_bottom;

    x_start = floor(y_orig * aspect_ratio_raw + 0.5);
    if (x_orig > x_start)
    	x_start = x_orig;
    if (800 > x_start)
    	x_start = 800;
    if (x_start % final_mod_x)
    	x_start += final_mod_x - x_start % final_mod_x;
    x_start += final_mod_x;

    y_start = floor(x_orig / aspect_ratio_raw + 0.5);
    if (y_orig > y_start)
    	y_start = y_orig;
    if (600 > y_start)
    	y_start = 600;
    if (y_start % final_mod_y)
    	y_start += final_mod_y - y_start % final_mod_y;
    y_start += final_mod_y;
    
    for (x = x_start; x >= final_mod_x; x -= 1)
    {
    	for (y = y_start; y >= final_mod_y; y -= 1)
    	{
	    scaled_crop_left = crop_left * (double) x / x_orig;
	    scaled_crop_right = crop_right * (double) x / x_orig;
	    scaled_crop_top = crop_top * (double) y / y_orig;
	    scaled_crop_bottom = crop_bottom * (double) y / y_orig;

	    x_img = x - scaled_crop_left - scaled_crop_right;
	    y_img = y - scaled_crop_top - scaled_crop_bottom;

	    x_new = x - ceil(scaled_crop_left) - ceil(scaled_crop_right);
	    y_new = y - ceil(scaled_crop_top) - ceil(scaled_crop_bottom);

	    x_mod = x_new % final_mod_x;
	    y_mod = y_new % final_mod_y;

	    scaled_crop_left = ceil(scaled_crop_left) + (x_mod >> 1);
	    scaled_crop_right = ceil(scaled_crop_right) + (x_mod >> 1);
	    scaled_crop_top = ceil(scaled_crop_top) + (y_mod >> 1);
	    scaled_crop_bottom = ceil(scaled_crop_bottom) + (y_mod >> 1);

	    temp = align_mod_x - (int) scaled_crop_left % align_mod_x;
	    if (temp != align_mod_x)
	    	scaled_crop_left += temp;
	    temp = align_mod_x - (int) scaled_crop_right % align_mod_x;
	    if (temp != align_mod_x)
	    	scaled_crop_right += temp;
	    temp = align_mod_y - (int) scaled_crop_top % align_mod_y;
	    if (temp != align_mod_x)
	    	scaled_crop_top += temp;
	    temp = align_mod_y - (int) scaled_crop_bottom % align_mod_y;
	    if (temp != align_mod_x)
	    	scaled_crop_bottom += temp;

	    if (align_mod_x == 1 && x_mod % 2)
	    	scaled_crop_right += 1;
	    if (align_mod_y == 1 && y_mod % 2)
	    	scaled_crop_bottom += 1;

	    x_new = x - scaled_crop_left - scaled_crop_right;
	    y_new = y - scaled_crop_top - scaled_crop_bottom;

	    if (x_new % final_mod_x || y_new % final_mod_y)
	    	continue;

    	    aspect_ratio_trial = (double) x / y;
    	    percent_error = (aspect_ratio_trial - aspect_ratio_raw) /
    			    aspect_ratio_raw * 100;

	    crop_x_error = (double) (x_img - x_new) / x_img * 100;
	    crop_y_error = (double) (y_img - y_new) / y_img * 100;

	    score = fabs(percent_error)  +
		    (crop_x_error + crop_y_error) / 2;

	    /* don't even bother keeping the crappy ones */
	    if (fabs(percent_error) <= 3.5 &&
	    	crop_x_error < 10 && crop_y_error < 10)
	    {

	    	res_array = (struct xy_res **) realloc(res_array,
				(num_entries + 1) * sizeof(struct xy_res *));
		res_array[num_entries] = calloc(1, sizeof(struct xy_res));

		res_array[num_entries]->x = x;
		res_array[num_entries]->y = y;
		res_array[num_entries]->x_new = x_new;
		res_array[num_entries]->y_new = y_new;
		res_array[num_entries]->scaled_crop_left = scaled_crop_left;
		res_array[num_entries]->scaled_crop_right = scaled_crop_right;
		res_array[num_entries]->scaled_crop_top = scaled_crop_top;
		res_array[num_entries]->scaled_crop_bottom = scaled_crop_bottom;
		res_array[num_entries]->percent_error = percent_error;
		res_array[num_entries]->crop_x_error = crop_x_error;
		res_array[num_entries]->crop_y_error = crop_y_error;
		res_array[num_entries]->score = score;

		num_entries++;
	    }
	}
    }

    /* keep only the best score for each resolution */
    qsort(res_array, num_entries, sizeof(struct xy_res *), compare_entries1);
    for (i = 0; i < num_entries; i++)
    {
	x_new = res_array[i]->x_new;
	y_new = res_array[i]->y_new;

    	if (x_new != old_x_new || y_new != old_y_new)
    	{
	    	res_array2 = (struct xy_res **) realloc(res_array2,
				(num_entries2 + 1) * sizeof(struct xy_res *));
		res_array2[num_entries2] = calloc(1, sizeof(struct xy_res));

		memcpy(res_array2[num_entries2], res_array[i],
		       sizeof(struct xy_res));

		num_entries2++;
    	}
    	
    	old_x_new = x_new;
    	old_y_new = y_new;
    }

#ifdef DEBUG
    qsort(res_array2, num_entries2, sizeof(struct xy_res *), compare_entries2);

    for (i = 0; i < num_entries2; i++)
    {
	printf("%4d %4d    %4d %4d    %2d %2d %2d %2d    %5.2f    %4.2f %4.2f    %f\n",
	       res_array2[i]->x_new, res_array2[i]->y_new,
	       res_array2[i]->x, res_array2[i]->y,
	       res_array2[i]->scaled_crop_left,
	       res_array2[i]->scaled_crop_right,
	       res_array2[i]->scaled_crop_top,
	       res_array2[i]->scaled_crop_bottom,
	       res_array2[i]->percent_error,
	       res_array2[i]->crop_x_error,
	       res_array2[i]->crop_y_error,
	       res_array2[i]->score);
    }
#endif



    /* Crop and then Resize */

    x = x_orig - crop_left - crop_right;
    y = y_orig - crop_top - crop_bottom;
    scaled_crop_left = crop_left;
    scaled_crop_right = crop_right;
    scaled_crop_top = crop_top;
    scaled_crop_bottom = crop_bottom;

    crop_x_error = 0.0;
    crop_y_error = 0.0;

    /* VirtualDub Lanczos resize crashes if input Y is odd */
    /* allow forcing of pre-resize mods for quirky resizers */
    /* only need to use this for mod 1 align resolutions */
    if (align_mod_x == 1 && pre_resize_mod_x > 1)
    {
	x_mod = x % pre_resize_mod_x;
	scaled_crop_left += x_mod >> 1;
	scaled_crop_right += x_mod >> 1;

	if (x_mod % 2)
	    scaled_crop_right += 1;

    	crop_x_error = x_mod / x * 100;
    	x -= x_mod;
    }
    if (align_mod_y == 1 && pre_resize_mod_y > 1)
    {
	y_mod = y % pre_resize_mod_y;

	scaled_crop_top += y_mod >> 1;
	scaled_crop_bottom += y_mod >> 1;

	if (y_mod % 2)
	    scaled_crop_bottom += 1;

    	crop_y_error = y_mod / y * 100;
    	y -= y_mod;
    }

    /* align_mod overrides any pre_resize settings */
    if (align_mod_x > 1 || align_mod_y > 1)
    {
    	int x_temp, y_temp;

	x_temp = x, y_temp = y;

	temp = align_mod_x - (int) scaled_crop_left % align_mod_x;
	if (temp != align_mod_x)
	{
    	    crop_x_error += temp / x_temp * 100;
    	    scaled_crop_left += temp;
    	    x -= temp;
    	}

	temp = align_mod_x - (int) scaled_crop_right % align_mod_x;
	if (temp != align_mod_x)
	{
    	    crop_x_error += temp / x_temp * 100;
    	    scaled_crop_right += temp;
    	    x -= temp;
    	}

	temp = align_mod_y - (int) scaled_crop_top % align_mod_y;
	if (temp != align_mod_y)
	{
    	    crop_y_error += temp / y_temp * 100;
    	    scaled_crop_top += temp;
    	    y -= temp;
    	}

	temp = align_mod_y - (int) scaled_crop_bottom % align_mod_y;
	if (temp != align_mod_y)
	{
    	    crop_y_error += temp / y_temp * 100;
    	    scaled_crop_bottom += temp;
    	    y -= temp;
    	}
    }

    y_scale = (x_orig / aspect_ratio_raw) / y_orig;
    aspect_ratio_cropped = (double) x / (y * y_scale);

    x_start = floor(y_orig * aspect_ratio_raw + 0.5);
    if (x_orig > x_start)
    	x_start = x_orig;
    if (800 > x_start)
    	x_start = 800;
    if (x_start % final_mod_x)
    	x_start += final_mod_x - x_start % final_mod_x;
    x_start += final_mod_x;

    y_start = floor(x_orig / aspect_ratio_raw + 0.5);
    if (y_orig > y_start)
    	y_start = y_orig;
    if (600 > y_start)
    	y_start = 600;
    if (y_start % final_mod_y)
    	y_start += final_mod_y - y_start % final_mod_y;
    y_start += final_mod_y;

    for (x_new = x_start; x_new >= final_mod_x; x_new -= final_mod_x)
    {
    	for (y_new = y_start; y_new >= final_mod_y; y_new -= final_mod_y)
    	{
    	    aspect_ratio_trial = (double) x_new / y_new;
    	    percent_error = (aspect_ratio_trial - aspect_ratio_cropped) /
    			    aspect_ratio_cropped * 100;

	    score = fabs(percent_error)  +
		    (crop_x_error + crop_y_error) / 2;

	    /* don't even bother keeping the crappy ones */
	    if (fabs(percent_error) <= 3.5 &&
	    	crop_x_error < 10 && crop_y_error < 10)
	    {
	    	res_array_cr = (struct xy_res **) realloc(res_array_cr,
				(num_entries_cr + 1) * sizeof(struct xy_res *));
		res_array_cr[num_entries_cr] = calloc(1, sizeof(struct xy_res));

		res_array_cr[num_entries_cr]->x = x_new;
		res_array_cr[num_entries_cr]->y = y_new;
		res_array_cr[num_entries_cr]->x_new = x_new;
		res_array_cr[num_entries_cr]->y_new = y_new;
		res_array_cr[num_entries_cr]->scaled_crop_left = scaled_crop_left;
		res_array_cr[num_entries_cr]->scaled_crop_right = scaled_crop_right;
		res_array_cr[num_entries_cr]->scaled_crop_top = scaled_crop_top;
		res_array_cr[num_entries_cr]->scaled_crop_bottom = scaled_crop_bottom;
		res_array_cr[num_entries_cr]->percent_error = percent_error;
		res_array_cr[num_entries_cr]->crop_x_error = crop_x_error;
		res_array_cr[num_entries_cr]->crop_y_error = crop_y_error;
		res_array_cr[num_entries_cr]->score = score;
		res_array_cr[num_entries_cr]->crop_then_resize_flag = 1;

		num_entries_cr++;
	    }
	}
    }

    /* keep only the best score for each resolution */
    qsort(res_array_cr, num_entries_cr, sizeof(struct xy_res *), compare_entries1);
    for (i = 0; i < num_entries_cr; i++)
    {
	x_new = res_array_cr[i]->x_new;
	y_new = res_array_cr[i]->y_new;

    	if (x_new != old_x_new || y_new != old_y_new)
    	{
	    	res_array_cr2 = (struct xy_res **) realloc(res_array_cr2,
				(num_entries_cr2 + 1) * sizeof(struct xy_res *));
		res_array_cr2[num_entries_cr2] = calloc(1, sizeof(struct xy_res));

		memcpy(res_array_cr2[num_entries_cr2], res_array_cr[i],
		       sizeof(struct xy_res));

		num_entries_cr2++;
    	}
    	
    	old_x_new = x_new;
    	old_y_new = y_new;
    }

#ifdef DEBUG
    qsort(res_array_cr2, num_entries_cr2, sizeof(struct xy_res *), compare_entries2);

    for (i = 0; i < num_entries_cr2; i++)
    {
	printf("%4d %4d    %4d %4d    %2d %2d %2d %2d    %5.2f    %4.2f %4.2f    %f *\n",
	       res_array_cr2[i]->x_new, res_array_cr2[i]->y_new,
	       res_array_cr2[i]->x, res_array_cr2[i]->y,
	       res_array_cr2[i]->scaled_crop_left,
	       res_array_cr2[i]->scaled_crop_right,
	       res_array_cr2[i]->scaled_crop_top,
	       res_array_cr2[i]->scaled_crop_bottom,
	       res_array_cr2[i]->percent_error,
	       res_array_cr2[i]->crop_x_error,
	       res_array_cr2[i]->crop_y_error,
	       res_array_cr2[i]->score);
    }
#endif



    /* Combine the two methods */
    res_array_combined = realloc(res_array_combined,
    		(num_entries2 + num_entries_cr2) * sizeof(struct xy_res *));
    for (i = 0; i < num_entries2; i++)
	res_array_combined[num_entries_combined++] = res_array2[i];
    for (i = 0; i < num_entries_cr2; i++)
	res_array_combined[num_entries_combined++] = res_array_cr2[i];

    /* keep only the best score for each resolution */
    qsort(res_array_combined, num_entries_combined,
    	  sizeof(struct xy_res *), compare_entries1);
    for (i = 0; i < num_entries_combined; i++)
    {
	x_new = res_array_combined[i]->x_new;
	y_new = res_array_combined[i]->y_new;

    	if (x_new != old_x_new || y_new != old_y_new)
    	{
	    	res_array_combined2 =
	    		(struct xy_res **) realloc(res_array_combined2,
			(num_entries_combined2 + 1) * sizeof(struct xy_res *));
		res_array_combined2[num_entries_combined2] =
			calloc(1, sizeof(struct xy_res));

		memcpy(res_array_combined2[num_entries_combined2],
			res_array_combined[i], sizeof(struct xy_res));

		num_entries_combined2++;
    	}
    	
    	old_x_new = x_new;
    	old_y_new = y_new;
    }

    qsort(res_array_combined2, num_entries_combined2,
    	  sizeof(struct xy_res *), compare_entries2);

    for (i = 0; i < num_entries_combined2; i++)
    {
	printf("%4d %4d    %4d %4d    %2d %2d %2d %2d    %5.2f    %4.2f %4.2f    %f",
	       res_array_combined2[i]->x_new, res_array_combined2[i]->y_new,
	       res_array_combined2[i]->x, res_array_combined2[i]->y,
	       res_array_combined2[i]->scaled_crop_left,
	       res_array_combined2[i]->scaled_crop_right,
	       res_array_combined2[i]->scaled_crop_top,
	       res_array_combined2[i]->scaled_crop_bottom,
	       res_array_combined2[i]->percent_error,
	       res_array_combined2[i]->crop_x_error,
	       res_array_combined2[i]->crop_y_error,
	       res_array_combined2[i]->score);
	if (res_array_combined2[i]->crop_then_resize_flag)
	    printf(" *");
	printf("\n");
    }

    /* suggested aspect ratios to try if things looked squished */
    /* probably because the DVD wasn't encoded at the correct aspect
       ratio ... */
    /* estimate an aspect ratio to try based on the cropped dimensions
       and the three common target dimensions: 4:3, 1.85:1, 2.35:1 */

    x = x_orig - crop_left - crop_right;
    y = y_orig - crop_top - crop_bottom;
    
    printf("\n");
    printf("Other aspect ratios to try if things still look squished:\n");
    printf(" (Estimated from the cropped dimensions and ideal film ratio)\n");
    printf("   4:3    %lf\n", x_orig / (((x / (4.0/3)) / y) * y_orig));
    printf("  16:9    %lf\n", x_orig / (((x / (16.0/9)) / y) * y_orig));
    printf("1.85:1    %lf\n", x_orig / (((x / 1.85) / y) * y_orig));
    printf("2.35:1    %lf\n", x_orig / (((x / 2.35) / y) * y_orig));

    return 0;
}
